import xarray as xr
from ..container import ensure_collection
from ..actuator_base import ExtractorBase


class Actuator(ExtractorBase):
    """Default feature extractor implementation.
    By default, the :py:meth:`transform` method does nothing
    and input data is just read.
    """

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Default feature extractor constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Estimator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping

        .. note::

          * The name of the variables to concatenate are given in the
            `'variables_to_concat'` entry of the container configuration.
            If this entry is empty, the name of the actuator result manager
            is used instead.
          * The dimension along which to concatenate must be given
            in the `'dim_to_concat'` entry of the container configuration.
        """
        # Variables to concatenate.
        variable_names = (
            ensure_collection(cfg.get('variables_to_concat'), set)
            or set([result_mng.result_name]))

        # Extractor initialization
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            variable_names=variable_names, **kwargs)

        #: Dimension along which to concatenate.
        self.dim_to_concat = self.cfg.get('dim_to_concat')

    def transform(self, data_src, **kwargs):
        """Default transform: return identical data source
        and prevent writing.

        :param data_src: Input multiple data source.
        :type data_src: :py:class:`..data_source.MultiDataSource`
        """
        self.log.info(
            '{} {} feature concatenation along {} dimension:'.format(
                self.component_mng.component_name, self.result_mng.result_name,
                self.dim_to_concat))

        # Get data (letting multiple data source manage variables)
        kwargs_data_src = kwargs.copy()
        kwargs_data_src.pop('variable_names', None)
        data_src.get_data(**kwargs_data_src)

        ds = {}
        for variable_name in self.variable_names:
            self.log.info(
                '  - Concatenating {} variable:'.format(variable_name))

            # Collect data for variable from all single data sources
            data = []
            for src_name, single_data_src in (
                    data_src.var_data_sources[variable_name].items()):
                self.log.info(
                    '    - Adding {} data source'.format(src_name))
                data.append(single_data_src[variable_name])

            # Concatenate
            ds[variable_name] = xr.concat(data, dim=self.dim_to_concat)

        return ds



"""ENTSOE API."""
import os
import numpy as np
import pandas as pd
import xarray as xr
from entsoe import EntsoePandasClient
from entsoe.mappings import PSRTYPE_MAPPINGS
from entsoe.exceptions import NoMatchingDataError
from ..geo import get_country_code
from ..data_source import DataSourceLoaderBase
from ..component import (DataSourceWithComponentsMixin,
                         parse_variable_component_args,
                         finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    #: Default source name.
    DEFAULT_SRC_NAME = 'entsoe'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

        self.sampling_conversion = {'P1Y': 'Y', 'PT60M': 'H'}

    @parse_variable_component_args
    def download(self, variable_component_names, **kwargs):
        """Download data.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from the
          `ENTSO-E Transparency Platform <https://transparency.entsoe.eu/>`.
        """
        # Define start and end dates (ISO format)
        start = pd.Timestamp(self.cfg['period_start'], tz=self.cfg['tz'])
        end = pd.Timestamp(self.cfg['period_end'], tz=self.cfg['tz'])
        svar = ', '.join(str(variable_name)
                         for variable_name in variable_component_names)
        if not self.cfg.get('no_verbose'):
            self.log.info(
                '{} variables from {} to {}:'.format(svar, start, end))

        # Get credentials and start ENTSO-E client
        cred = self.med.cfg.get_credentials(
            self.DEFAULT_SRC_NAME, ['security_token'])
        client = EntsoePandasClient(api_key=cred['security_token'])

        # Get zone-code mapping and zone-countrycode mapping
        place_code, place_country_code, place_lookup = self._get_place_code(
            **kwargs)

        # Loop over variables
        for variable_name, component_names in variable_component_names.items():
            # Define query
            src_variable_name = self.cfg['variable_names'][variable_name]
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))
            query = getattr(client, 'query_{}'.format(src_variable_name))

            # Manage wind component as onshore plus offshore
            component_name = 'wind'
            component_to_download_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are downloaded
                component_to_download_names.update(
                    ['wind-onshore', 'wind-offshore'])

                # Remove capacity factor from components to download
                component_to_download_names.discard(component_name)

            # Loop over components
            for component_name in component_to_download_names:
                src_component_name = self.cfg['component_names'][
                    component_name]
                psr_type = [k for k, v in PSRTYPE_MAPPINGS.items()
                            if v == src_component_name][0]
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {} -> {} -> {}:'.format(
                        component_name, src_component_name, psr_type))
                # Loop over zones
                without_zones = True
                for place_name, code in place_code.items():
                    country_code = place_country_code[place_name]
                    if not self.cfg.get('no_verbose'):
                        self.log.info('--- {}'.format(place_name))
                    try:
                        # Try with zones
                        df = query(
                            code, start=start, end=end, psr_type=psr_type,
                            lookup_bzones=place_lookup[place_name])
                    except TypeError:
                        if without_zones:
                            # Get dataset without zones instead
                            if place_lookup[place_name]:
                                self.log.warning(
                                    '{} dataset not zonal: downloading '
                                    'for country instead'.format(
                                        variable_name))
                            df = query(country_code, start=start, end=end,
                                       psr_type=psr_type)
                            without_zones = False
                        else:
                            # Otherwise break zones loop
                            break
                    except NoMatchingDataError:
                        if not self.cfg.get('no_verbose'):
                            self.log.warning(
                                'No mathcing data found for {} {} {}.'
                                ' Saving empty data frame.'.format(
                                    place_name, component_name, variable_name))
                        df = pd.DataFrame(columns=[component_name])

                    # Adjust names
                    df.index.name = 'time'
                    df.columns = [component_name]

                    # Save locally
                    filepath = self._get_download_filepath(
                        variable_name, component_name, place_name, code,
                        **kwargs)
                    df.to_csv(filepath)

        # Return names of downloaded variables
        return variable_component_names

    @parse_variable_component_args
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: :py:class:`xarray.Dataset()`
        """
        # Get zone-code mapping
        place_code, _, _ = self._get_place_code(**kwargs)

        # Loop over datasets
        ds = {}
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Manage wind component as onshore plus offshore
            component_name = 'wind'
            component_to_load_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are loaded
                component_to_load_names.update(
                    ['wind-onshore', 'wind-offshore'])

                # Remove capacity factor from components to load
                component_to_load_names.discard(component_name)

            # Loop over components
            da = None
            for component_name in component_to_load_names:
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {}:'.format(component_name))
                # Loop over zones
                da_comp = None
                for place_name, code in place_code.items():
                    if not self.cfg.get('no_verbose'):
                        self.log.info('--- {}'.format(place_name))
                    # Read downloaded data
                    filepath = self._get_download_filepath(
                        variable_name, component_name, place_name, code,
                        **kwargs)
                    df = pd.read_csv(filepath, index_col=0, parse_dates=True)

                    # Convert index timezone to UTC
                    df.index = pd.to_datetime(
                        df.index, utc=True).tz_convert(None)

                    # Convert to DataArray
                    da_zone = xr.DataArray(
                        df, dims=('time', 'component')).expand_dims(
                            'region').assign_coords(region=[place_name])

                    area = self.med.geo_src.place_area[place_name]
                    if (area == 'Italy') and (variable_name == 'generation'):
                        # Set inconsistent values to NaN
                        da_zone.values.setflags(write=True)
                        start_na = pd.Timestamp('2016-10-26T00:00')
                        end_na = pd.Timestamp('2016-11-01T00:00')
                        da_zone.loc[{'time': slice(start_na, end_na)}] = np.nan

                    # Concatenate
                    da_comp = (da_zone if da_comp is None else
                               xr.concat([da_comp, da_zone], dim='region'))

                # Concatenate
                da = (da_comp if da is None else
                      xr.concat([da, da_comp], dim='component'))

            # Add total wind component if needed
            component_name = 'wind'
            if component_name in component_names:
                # Treat NaNs as zero unless in both components
                da_on = da.sel(component='wind-onshore', drop=True)
                da_off = da.sel(component='wind-offshore', drop=True)
                na = da_on.isnull() & da_off.isnull()
                da_wind = (da_on.fillna(0.) + da_off.fillna(0.)).where(
                    ~na, np.nan)
                da_wind = da_wind.expand_dims('component').assign_coords(
                    component=[component_name])
                da = xr.concat([da, da_wind], dim='component')

            # Add variable to dataset
            ds[variable_name] = da

        return ds

    def _get_place_code(self, **kwargs):
        """Get zone to zone-code mapping and zone to country-code mapping.

        :returns: Zone to zone code, zone to country code and
          zone to lookup flag mappings.
        :rtype: :py:class:`tuple` of

          * mapping of :py:class:`str` to :py:class:`str`
          * mapping of :py:class:`str` to :py:class:`str`
          * mapping of :py:class:`str` to :py:class:`bool`

        .. seealso:: :py:meth:`..geo.get_country_code`.
        """
        # Get place names for source
        src_place_regions, _ = (
            self.med.geo_src.get_place_regions_for_source(
                self.name, **kwargs))

        place_code = {}
        place_country_code = {}
        place_lookup = {}
        for place_name, region_names in src_place_regions.items():
            # Get place (zone) area (country)
            area = self.med.geo_src.place_area[place_name]

            # Get country code
            place_country_code[place_name] = get_country_code(
                area, code='alpha-2')

            if area == place_name:
                # Assign country-code to place (country)
                place_code[area] = place_country_code[place_name]

                # Do not look for building zone for this place
                place_lookup[area] = False
            else:
                # Assign code to place (zone)
                place_code[place_name] = '{}-{}'.format(
                    place_country_code[place_name], region_names[0])

                # Look for building zone for this place
                place_lookup[place_name] = True

        return place_code, place_country_code, place_lookup

    def _get_download_filepath(
            self, variable_name, component_name, place_name, code, **kwargs):
        """Get downloaded dataset filepath.

        :param variable_name: Dataset name.
        :param component_name: Component name.
        :param place_name: Zone/country name.
        :param code: Zone/country code.
        :type variable_name: str
        :type component_name: str
        :type place_name: str
        :type code: str

        :returns: Filepath
        :rtype: str
        """
        src_dir = self.med.cfg.get_external_data_directory(self)
        src_dir = os.path.join(src_dir, variable_name)
        os.makedirs(src_dir, exist_ok=True)
        filename = '{}_{}_{}_{}{}.csv'.format(
            self.name, variable_name, component_name, code,
            self.get_postfix(**kwargs))
        filepath = os.path.join(src_dir, filename)

        return filepath

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(
            self.cfg['period_start'], self.cfg['period_end'])

        return postfix



"""Mediator-related definitions."""
import os
import collections
import logging
from warnings import warn
from .config import Config, import_module_from_cfg
from .component import ComponentManager
from .data_source import MultiDataSource
from .geo import MultiGeoDataSource
from .container import Container, ensure_collection, load_config


class Mediator(Container, collections.MutableMapping):
    """Models mediator."""

    def __init__(self, cfg, initialize=True):
        """Constructor setting configuration and task manager.

        :param cfg: Configuration or a file path from which to read
          configuration.
        :param initialize: Whether to initialize problem, or to stop
          after loading mediator configuration. Default is `True`.
        :type cfg: mapping or :py:class:`str`
        :type initialize: bool
        """
        # Load configuration
        loaded_cfg = cfg if isinstance(cfg, Config) else Config(cfg)

        #: Data-sources mapping.
        self.data_sources = collections.OrderedDict()

        #: Geographic configuration.
        self.geo_cfg = None
        #: Geographic data-source configuration.
        self.geo_src = None

        #: Component-managers mapping.
        self.component_managers = None

        #: Optimizer.
        self.optimizer = None

        #: Logger.
        self.log = logging.getLogger(loaded_cfg['project_name'])

        # Define as container without parent
        super(Mediator, self).__init__(self, 'mediator', cfg=loaded_cfg)

        if initialize:
            # Initialize
            self.initialize()

    def initialize(self):
        """Initialize mediator."""
        # Set-up logger
        _init_logger(self.log, self.cfg)

        if not self.cfg.get('no_verbose'):
            self.log.info('*** INITIALIZING MEDIATOR FOR {} ***'.format(
                self.cfg.get('project_name')))

        # Inject geography in mediator as data source
        self._init_geo()

        # Inject all components in mediator
        self._init_component_managers()

        # Inject optimizer
        self._init_optimizer()

    def add_data_source(self, src_vars, parent=None, isgeo=False,
                        **kwargs):
        """ Add single or multiple data source depending on whether
        :py:obj:`src_dict` has keys.

        :param src_vars: Mapping with multiple
          (source name, variable_names collection)
          (key, value) pairs, sequence of source names, source name.
        :param parent: Container for which to load the source.
          Default is `None`.
        :param isgeo: Whether data source is :py:class`.geo.GeoDataSourceBase`.
          Default is `False`.
        :type src_vars: Mutable mapping, collection or string.
        :type container: :py:class:`.container.Container`
        :type isgeo: bool

        :returns: Data source.
        :rtype: :py:class:`.data_source.DataSourceBase`

        .. seealso:: :py:meth:`.actuator_base.ActuatorBase.add_data_sources`
        """
        if ((not isinstance(src_vars, str)) and
            isinstance(src_vars, collections.Collection) and
                (len(src_vars) > 1)):
            data_src = self.med.add_multi_data_source(
                src_vars, parent=parent, isgeo=isgeo, **kwargs)
        else:
            if isinstance(src_vars, collections.Mapping):
                src_name, variable_names = src_vars.copy().popitem()
            else:
                # Try to get variables from keyword arguments,
                # otherwise, leave None
                variable_names = kwargs.pop('variable_names', None)
                src_name = str(*ensure_collection(src_vars))
            data_src = self.med.add_single_data_source(
                src_name, variable_names=variable_names, parent=parent,
                **kwargs)

        return data_src

    def add_single_data_source(self, name, variable_names=None, parent=None,
                               **kwargs):
        """Initialize a data source, inject to mediator,
        and return it as well.

        :param name: Data source name.
        :param variable_names: Data source variables. Default is `None`,
          in which case variables should be defined after construction
          by data source or by user.
        :param parent: Container for which to add data source.
          Default is `None`.
        :type name: str
        :type variable_names: (collection of) :py:class:`str`
        :type parent: :py:class:`.container.Container`

        :returns: (Multiple) data source.
        :rtype: :py:class:`.data_source.DataSourceBase`

        .. seealso:: :py:meth:`add_data_source`
        """
        # Verify that data source not already added by other containers
        if name not in self.data_sources:
            # Add data source configuration
            cfg_src = load_config(self, name, parent)

            if cfg_src is None:
                # Raise exception if configuration not found
                raise ValueError('No {} data source injected: no '
                                 'configuration found'.format(name))

            # Import data-source module
            mod = import_module_from_cfg(cfg_src, name=name)

            if mod is None:
                raise ValueError('Could not import module for '
                                 '{} data source'.format(name))

            # Create data source and inject it to mediator
            self.data_sources[name] = mod.DataSource(
                self, name, cfg=cfg_src, variable_names=variable_names,
                **kwargs)

            if not self.cfg.get('no_verbose'):
                self.log.info(
                    '{} data source injected to mediator'.format(name))
        else:
            # Add variable names to existing data source, if needed
            if variable_names is not None:
                self.data_sources[name].update_variables(variable_names)

        return self.data_sources[name]

    def add_multi_data_source(self, src_vars, parent=None, isgeo=False):
        """Initialize multiple data source, inject to mediator,
        and return it as well.

        :param src_vars: Mapping with multiple
          (source name, variable_names list)
          (key, value) pairs, sequence of source names..
        :param parent: Container for which to add data source.
          Default is `None`.
        :param isgeo: Whether data source is :py:class`.geo.GeoDataSourceBase`.
          Default is `False`.
        :type src_vars: mapping, or collection.
        :type parent: :py:class:`.container.Container`
        :type isgeo: bool

        :returns: Multiple data source.
        :rtype: :py:class:`.data_source.MultiDataSource` or
          :py:class:`.data_source.MultiGeoDataSource`

        .. seealso:: :py:meth:`add_data_source`,
          :py:meth:`add_single_data_source`
        """
        if not isinstance(src_vars, collections.Mapping):
            # If a collection of source names is given, transform it
            # to a mapping from source names to None (variable_names)
            src_vars = {src_name: None for src_name in src_vars}

        data_sources = collections.OrderedDict()
        for src_name, variable_names in src_vars.items():
            # Add data source to mediator
            self.add_single_data_source(
                src_name, variable_names=variable_names, parent=parent)

            # Add data source to multi-data source dictionary
            data_sources[src_name] = self.data_sources[src_name]

        # Verify that multi data source not already added by other containers
        multi_src_name = MultiDataSource.get_name(data_sources)
        if multi_src_name not in self.data_sources:
            # Create multiple data source from data sources dictionary
            multi_data_src = (MultiGeoDataSource(self, data_sources) if isgeo
                              else MultiDataSource(self, data_sources))

            # Inject multiple data source to mediator
            self.data_sources[multi_src_name] = multi_data_src
        else:
            # Get existing multiple data source
            multi_data_src = self.data_sources[multi_src_name]

            # Update multiple data source with potential new data sources
            multi_data_src.update_data_sources(data_sources)

        return multi_data_src

    def _init_geo(self, **kwargs):
        """Initialize geographic data source(s).

        .. note:: Only data sources associated to areas present in
          the 'component_managers_per_area' entry of the mediator configuration
          are added.
        """
        self.geo_cfg = load_config(self, 'geo')
        if self.geo_cfg is not None:
            # Add data sources for areas required by components
            src_names = set([self.geo_cfg['data'][area] for area in self.cfg[
                'component_managers_per_area']])
            self.add_data_source(src_names, isgeo=True)

            # Assign convenience member pointing to (multiple) geo data source
            src_names = ensure_collection(src_names)
            multi_src_name = (MultiDataSource.get_name(src_names)
                              if len(src_names) > 1 else src_names[0])
            self.geo_src = self.data_sources[multi_src_name]
        else:
            self.log.warning('No geographic data source provided: skipping')

    def _init_component_managers(self, **kwargs):
        """Initialize components."""
        # Get area-component managers mapping
        area_component_managers = self.cfg.get('component_managers_per_area')

        if area_component_managers is not None:
            # Initialize component-managers mapping
            self.component_managers = collections.OrderedDict()

            # Loop over area-component managers pairs
            for area, cfg_component_managers in (
                    area_component_managers.items()):
                if not self.cfg.get('no_verbose'):
                    scmp = ', '.join(cfg_component_managers.keys())
                    self.log.info(
                        'Injecting {} component managers for {} to '
                        'mediator'.format(scmp, area))

                # Loop over component manager-result managers pairs
                for component_mng_name, result_mng_names in (
                        cfg_component_managers.items()):
                    # Add component manager
                    self.component_managers[component_mng_name] = (
                        ComponentManager(self, component_mng_name, area=area,
                                         result_mng_names=result_mng_names))
        else:
            self.log.warning('No component manager provided: skipping')

    def _init_optimizer(self, **kwargs):
        """Initialize optimizer to associate to the given mediator."""
        # Add optimizer configuration
        opt_name = self.cfg.get('optimizer')
        if opt_name is not None:
            name = self.cfg['optimizer']
            cfg_opt = load_config(self, name)

            # Import data-source module
            mod = import_module_from_cfg(cfg_opt, name=name)

            # Add data source to mediator
            self.optimizer = mod.Optimizer(self, cfg=cfg_opt, **kwargs)

            if not self.cfg.get('no_verbose'):
                self.log.info('{} injected to mediator'.format(
                    self.optimizer.name))
        else:
            self.log.warning('No optimizer provided: skipping')

    def get_component_names_per_variable_for_source(
            self, src_name, variable_names=None, component_names=None,
            **kwargs):
        """Get component names per variable for data source.

        :param src_name: Data-source name.
        :param variable_names: Variable names.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are returned.
        :param component_names: Component names.
          Default is `None`, in which case all mediator components
          that have :py:obj:`self` as data source are returned.
        :type src_name: str
        :type variable_names: (collection of) :py:class:`str`
        :type component_names: (collection of) :py:class:`str`

        :returns: Component names.
        :rtype: set
        """
        # Ensure collection
        component_names = ensure_collection(component_names, set)

        # Get set of component managers
        if component_names is None:
            # From mediator component managers
            component_managers = self.component_managers
        else:
            # From mediator component managers whose component names are in
            # the given component names
            component_managers = collections.OrderedDict()
            for name, component_mng in self.component_managers.items():
                if component_mng.component_name in component_names:
                    component_managers[name] = component_mng

        d = collections.OrderedDict()
        for component_mng in component_managers.values():
            if src_name in component_mng.data_sources:
                if variable_names is None:
                    # Add variables associated to component results
                    for result_mng in component_mng.result_managers.values():
                        d.setdefault(result_mng.result_name, set()).add(
                            component_mng.component_name)
                else:
                    # Add all variables for component
                    for variable_name in variable_names:
                        d.setdefault(variable_name, set()).add(
                            component_mng.component_name)

        return d

    def __getitem__(self, component_mng_name):
        """Get component from :py:attr:`components`.

        :param component_mng_name: Component-manager name.
        :type component_mng_name: str

        :returns: Component manager.
        :rtype: :py:class:`.component.ComponentManager`
        """
        return self.component_managers[component_mng_name]

    def get(self, component_mng_name, default=None):
        """Get component from :py:attr`data`.

        :param component_mng_name: Component-manager name.
        :param default: Default value. Default is `None`.
        :type component_mng_name: str
        :type default: :py:class:`.component.ComponentManager`

        :returns: component.
        :rtype: :py:class:`.component.ComponentManager`
        """
        return self.component_managers.get(component_mng_name, default)

    def __setitem__(self, component_mng_name, component_mng):
        """Set component in :py:attr:`data`.

        :param component_mng_name: Component-manager name.
        :param component_mng: Component manager to set.
        :type component_mng_name: str
        :type component_mng: :py:class:`.component.ComponentManager`
        """
        self.component_managers[component_mng_name] = component_mng

    def __contains__(self, component_mng_name):
        """Test if component manager in :py:attr:`component_managers`.

        :param component_mng_name: component name.
        :type component_mng_name: str
        """
        return component_mng_name in self.component_managers

    def __delitem__(self, component_mng_name):
        """Remove component manager from :py:attr:`component_managers`.

        :param component_mng_name: Component-manager name.
        :type component_mng_name: str
        """
        del self.component_managers[component_mng_name]

    def __iter__(self):
        """Iterate :py:attr:`component_managers` mapping."""
        return iter(self.component_managers)

    def __len__(self):
        """Number of component managers."""
        return len(self.component_managers)


def _init_logger(log, cfg):
    """Initialize logger.

    :param log: Logger.
    :param cfg: Logger configuration.
    :type log: :py:class:`logging.Logger`
    :type cfg: mapping
    """
    log_path = cfg.get('log_path')
    log_level = cfg.get('log_level', 'INFO')
    log_fmt0 = ('%(asctime)s/%(name)s/%(module)s.%(funcName)s:%(lineno)d/'
                '%(levelname)s: %(message)s')
    log_fmt = cfg.get('log_fmt', log_fmt0)
    capture_warnings = cfg.get('capture_warnings', True)

    # Configure logger
    log.setLevel(log_level)

    # Create file or stream handler
    if log_path is not None:
        # If log_path given, log to file instead of stream
        log_path = os.path.join(*ensure_collection(log_path))
        msg = ('Setting loggging to {}. '
               'Open it for further information'.format(log_path))
        warn(msg)
        # Make sure that log-file directory exists
        os.makedirs(os.path.dirname(log_path), exist_ok=True)
        hdlr = logging.FileHandler(log_path, mode='a')
    else:
        hdlr = logging.StreamHandler()

    # Add formatter and set level
    formatter = logging.Formatter(log_fmt)
    hdlr.setFormatter(formatter)
    hdlr.setLevel(log_level)

    # Add handler to root logger if needed
    if not log.hasHandlers():
        log.addHandler(hdlr)

    # Add handler to warning logger if needed
    if capture_warnings:
        logging.captureWarnings(capture_warnings)
        warn_log = logging.getLogger('py.warnings')
        if not warn_log.hasHandlers():
            warn_log.addHandler(hdlr)

    # Add handler to matplotlib logger if needed
    mpl_log = logging.getLogger('matplotlib')
    if not mpl_log.hasHandlers():
        mpl_log.addHandler(hdlr)



"""gov.uk DUKES API for Great Britain."""
import os
import requests
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (DataSourceWithComponentsMixin,
                         parse_variable_component_args,
                         finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    """gov.uk DUKES data source."""
    #: Default source name.
    DEFAULT_SRC_NAME = 'gov_uk_dukes'

    #: Area: `'United Kingdom of Great Britain and Northern Ireland'`
    AREA = 'United Kingdom of Great Britain and Northern Ireland'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    @parse_variable_component_args
    def download(self, variable_component_names, **kwargs):
        """Download data.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from the `gov.uk DUKES's Chapter 6 <https://www.gov.uk/government/statistics/renewable-sources-of-energy-chapter-6-digest-of-united-kingdom-energy-statistics-dukes/>
        """
        # Loop over variables
        downloaded_files = []
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Loop over components
            for component_name in component_names:
                # Get URL
                url, filepath = self._get_url_filepath(
                    variable_name, component_name, **kwargs)

                # Only download once
                if filepath not in downloaded_files:
                    if not self.cfg.get('no_verbose'):
                        self.log.info('-- {} from {} to {}'.format(
                            component_name, url, filepath))

                    # Request and raise exception if needed
                    response = requests.get(url)
                    response.raise_for_status()

                    # Write XLS
                    with open(filepath, 'wb') as f:
                        f.write(response.content)

                    # Prevent further downloads
                    downloaded_files.append(filepath)

        # Return names of downloaded variables
        return variable_component_names

    @parse_variable_component_args
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: :py:class:`xarray.Dataset()`
        """
        # Get place names for source
        src_place_regions, _ = (
            self.med.geo_src.get_place_regions_for_source(
                self.name, self.AREA, **kwargs))

        # Loop over variables
        ds = xr.Dataset()
        loaded_files = []
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Loop over components
            da = None
            for component_name in component_names:
                # Get filepath
                _, filepath = self._get_url_filepath(
                    variable_name, component_name, **kwargs)

                # Only download once
                if filepath not in loaded_files:
                    if not self.cfg.get('no_verbose'):
                        self.log.info('-- {} from {}'.format(
                            component_name, filepath))

                    # Read data frame
                    read_excel_kwargs, icols = self._get_read_excel_kwargs(
                        variable_name, component_name, **kwargs)
                    df = pd.read_excel(filepath, **read_excel_kwargs)

                    # Transpose and select columns
                    df = df.transpose().iloc[:, icols]

                    # Define time index
                    df.index = pd.DatetimeIndex(
                        ('{}-12-31'.format(y) for y in df.index), freq='A-DEC')
                    df.index.name = 'time'

                    # Select period
                    df = df.loc[self.cfg['start_date']:self.cfg['end_date']]

                    # Prevent further downloads
                    loaded_files.append(filepath)

                # Select component
                df_comp = df[self.cfg['component_names'][
                    component_name]].astype(float)

                # Convert to data array
                da_comp = xr.DataArray(df_comp, dims=('time',)).expand_dims(
                    'component').assign_coords(component=[component_name])

                # Add component to data array
                da = (da_comp if da is None else
                      xr.concat([da, da_comp], dim='component'))

            # Add area as region coordinate
            da = da.expand_dims('region').assign_coords(
                region=list(src_place_regions))

            # Add to dataset
            ds[variable_name] = da

        return ds

    def _get_url_filepath(self, variable_name, component_name, **kwargs):
        """Get URL for query and filepath.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :type variable_name: str
        :type component_name: str

        :returns: Tuple containing URL and filepath.
        :rtype: :py:class:`tuple` with two :py:class:`str`
        """
        read_excel_kwargs, _ = self._get_read_excel_kwargs(
            variable_name, component_name, **kwargs)

        # Get URL and filename
        if component_name in ['wind', 'wind-onshore', 'wind-offshore',
                              'marine', 'pv', 'hydro', 'bio']:
            if variable_name == 'capacity_factor':
                filename = '{}_{}.xls'.format(
                    'DUKES', read_excel_kwargs['sheet_name'])
                url = '{}/{}/{}'.format(
                    self.cfg['host'], '822656', filename)

        # Get filepath
        src_dir = self.med.cfg.get_external_data_directory(self)
        filepath = os.path.join(src_dir, filename)

        return url, filepath

    def _get_read_excel_kwargs(self, variable_name, component_name, **kwargs):
        """Get keyword arguments for :py:meth:`pandas.read_excel` and columns slice.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :type variable_name: str
        :type component_name: str

        :returns: Tuple with keywoard arguments and column slice.
        :rtype: :py:class:`tuple` with :py:class:`dict` and :py:class:`slice`
        """
        if component_name in ['wind', 'wind-onshore', 'wind-offshore',
                              'marine', 'pv', 'hydro', 'bio']:
            if variable_name == 'capacity_factor':
                kwargs = {'sheet_name': '6.5', 'index_col': 0, 'header': 3,
                          'skipfooter': 7}
                icols = slice(20, None)

        return kwargs, icols

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(
            self.cfg['start_date'], self.cfg['end_date'])

        return postfix



"""Modifier to weight dataset by grid area."""
import numpy as np
import xarray as xr
import iris
from ..actuator_base import ExtractorBase


class Actuator(ExtractorBase):
    """Actuator to weight a gridded dataset by grid-boxes areas."""

    def transform(self, ds=None, **kwargs):
        """Apply modifier to data source.

        :param ds: Dataset to transform.
        :type ds: :py:class:`xarray.Dataset`

        :returns: Modified dataset.
        :rtype: Same as :py:obj:`data_src.data`
        """
        for da_name, da in ds.items():
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Area weighting {} {}'.format(
                        self.component_mng.name, da_name))

            # Get cube from array
            cube = da.to_iris()
            try:
                # Try to get area from existing bounds
                area = iris.analysis.cartography.area_weights(cube)
            except ValueError:
                # Get area from guessed bounds
                try:
                    cube.coord('longitude').guess_bounds()
                    cube.coord('latitude').guess_bounds()
                except iris.exceptions.CoordinateNotFoundError:
                    cube.coord('grid_longitude').guess_bounds()
                    cube.coord('grid_latitude').guess_bounds()
                area = iris.analysis.cartography.area_weights(cube)

            # Convert to array
            da_area = xr.DataArray(area, coords=da.coords)

            if self.cfg.get('normalize'):
                # Normalize by mean area
                da_area /= da_area.sum() / np.prod(da_area.shape)

            # Weight by area
            ds[da_name] *= da_area

        return ds

    def get_extractor_postfix(self, **kwargs):
        """Get extractor postfix.

        returns: Postfix.
        rtype: str
        """
        postfix = '{}_weighted'.format(
            super(Actuator, self).get_extractor_postfix(**kwargs))

        postfix += '_norm' * self.cfg['normalize']

        return postfix



"""Energimyndigheten API for Sweden."""
import pandas as pd
import xarray as xr
from pkg_resources import resource_filename
from ..data_source import DataSourceLoaderBase
from ..component import (
    DataSourceWithComponentsMixin, parse_variable_component_args,
    compute_capacity_factor, finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    #: Default source name.
    DEFAULT_SRC_NAME = 'energimyndigheten'

    #: Area: `'Sweden'`
    AREA = 'Sweden'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def download(self, *args, **kwargs):
        """Convenience function to warn that Energimyndigheten data is not
        downloaded but loaded from package data.

        :returns: Downloaded variable names.
        :rtype: :py:class:`set` of :py:class:`str`

        .. note:: The data was downloaded from the
          `Energimyndigheten Statistics Database <https://pxexternal.energimyndigheten.se/pxweb/en/Vindkraftsstatistik/>`.
        """
        self.log.warning(
            '{} is not to be downloaded. It is instead directly loaded from '
            'the package data.'.format(self.DEFAULT_SRC_NAME))

        return {'generation', 'capacity', 'capacity_factor'}

    @parse_variable_component_args
    @compute_capacity_factor
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: dict
        """
        # Get places and regions
        src_place_regions, _ = self.med.geo_src.get_place_regions_for_source(
            self.name, self.AREA, **kwargs)
        src_place_names = list(src_place_regions)

        ds = {}
        ddf_allvar = {}
        for variable_name, component_names in variable_component_names.items():
            # Get column for variable
            col = self.cfg['variable_names'][variable_name]

            # Manage wind component as onshore only
            component_name = 'wind'
            component_to_load_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are loaded
                component_to_load_names.add('wind-onshore')

                # Remove capacity factor from components to load
                component_to_load_names.discard(component_name)

            da = None
            for component_name in component_to_load_names:
                # Try to get previously-read component data
                df_allvar_comp = ddf_allvar.get(component_name)

                if df_allvar_comp is None:
                    # Read data
                    resource_name = self.get_download_resource_name(
                        component_name, **kwargs)
                    df_allvar_comp = pd.read_csv(
                        resource_filename(__name__, resource_name),
                        **self.cfg['read_csv_kwargs'])

                    # Save data
                    ddf_allvar[component_name] = df_allvar_comp

                # Select variable
                df_comp = df_allvar_comp[col].unstack()

                # Get time index with an additional year to count days in years
                # starting the first day of the year to count days
                idx = df_comp.index
                time = pd.DatetimeIndex((
                    '{}-01-01'.format(str(y)) for y in (
                        list(idx) + [int(idx[-1]) + 1])))

                # Aggregate in zones
                df_comp = self._aggregate_in_zones(
                    df_comp, time[:-1], src_place_names, src_place_regions,
                    **kwargs)

                # Replace time to 'A-DEC'
                df_comp.index = pd.date_range(
                    start=df_comp.index[0], end=df_comp.index[-1] +
                    pd.Timedelta('366 days'), freq='A-DEC')

                # Convert to data array
                da_comp = xr.DataArray(
                    df_comp, dims=['time', 'region'],
                    name=variable_name).expand_dims('component').assign_coords(
                        component=[component_name])

                # Add total wind component as onshore wind if needed
                component_name_add = 'wind'
                if component_name_add in component_names:
                    coord_component = da_comp.indexes['component'].tolist()
                    idx = coord_component.index('wind-onshore')
                    coord_component[idx] = component_name_add
                    da_comp['component'] = coord_component

                # Add component to variable in dataset
                da = (da_comp if da is None else
                      xr.concat([ds[variable_name], da_comp], dim='component'))

            ds[variable_name] = da

        return ds

    def get_download_resource_name(self, component_name, **kwargs):
        """Get package resource name for data.

        :returns: Resource name.
        :rtype: str
        """
        filename = '{}_production_{}{}.csv'.format(
            component_name, self.cfg['level'], self.get_postfix(**kwargs))
        resource_name = '../data/{}/{}/{}'.format(
            self.AREA, self.DEFAULT_SRC_NAME, filename)

        return resource_name

    def _get_capacity_factor(self, ds, **kwargs):
        """Get capacity factor from generation and capacity.

        :param ds: Dataset containing generation and capacity.
        :type ds: mapping

        :returns: Capacity factor.
        :rtype: :py:class:`xarray.DataArray`
        """
        da_gen = ds['generation']

        # Get number of days in month
        time = da_gen.indexes['time']
        start = pd.Timestamp(time[0])
        end = pd.Timestamp(time[-1]) + pd.Timedelta('366 days')
        time_plus = pd.date_range(start=start, end=end, freq='Y')
        ndays = (time_plus[1:] - time_plus[:-1]).days
        da_ndays = xr.DataArray(ndays, coords=[('time', time)])
        da_nhours = da_ndays * 24

        # Get capacity factor
        da = da_gen / (ds['capacity'] * da_nhours)

        return da

    def _aggregate_in_zones(
            self, ddf, time, src_place_names, src_place_regions, **kwargs):
        """Aggregate data in zones.

        :param ddf: Dataset containing variables.
        :param time: Time index.
        :param src_place_names: Zone/region names
        :param src_place_regions: Zone/region to regions mapping.
        :type ddf: mapping from :py:class:`str` to :py:class:`pandas.DataFrame`
        :type time: :py:class:`pandas.DatetimeIndex`
        :type src_place_names: collection
        :type src_place_regions: mapping from :py:class:`str` to collection

        :returns: Dataset containing aggregated variables.
        :rtype: mapping from :py:class:`str` to :py:class:`pandas.DataFrame`
        """
        ddf_zones = {}
        for variable_name, df in ddf.items():
            # Assign time index
            df.index = time

            # Aggregate into zones
            df_zones = pd.DataFrame(
                0., index=df.index, columns=src_place_names)
            for place_name, region_names in src_place_regions.items():
                df_zones[place_name] = df[region_names].sum('columns')

            # Add
            ddf_zones[variable_name] = df_zones

        return ddf_zones

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(
            self.cfg['first_year'], self.cfg['last_year'])

        return postfix



"""GME API."""
import os
import requests
import zipfile
from io import BytesIO
import pandas as pd
import xarray as xr
from ..container import ensure_collection
from ..data_source import DataSourceLoaderBase
from ..component import (DataSourceWithComponentsMixin,
                         parse_variable_component_args,
                         finalize_dataset)


COMPONENT_NAME = 'demand'
VARIABLE_NAME = 'demand'


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    #: Default source name.
    DEFAULT_SRC_NAME = 'gme'

    #: Area: `'Italy'`.
    AREA = 'Italy'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping

        .. note:: The GME data is downloaded from the
          `GME website <http://www.gestoremercatienergetici.org>`_.
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def download(self, **kwargs):
        """Download GME data source.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`
        """
        years = range(int(self.cfg['first_year']),
                      int(self.cfg['last_year']) + 1)
        for year in years:
            # Download file
            url = self._get_url(year, **kwargs)
            if not self.cfg.get('no_verbose'):
                self.log.info('- Year {} from {}'.format(year, url))

            # Request and raise exception if needed
            response = requests.get(url)
            response.raise_for_status()

            # Get zip
            zip_ref = zipfile.ZipFile(BytesIO(response.content))

            # Get filepath with extension
            filepath = self._get_download_filepath(year, **kwargs)
            extension = os.path.splitext(zip_ref.filelist[0].filename)[-1]
            filepath = '{}{}'.format(filepath, extension)

            # Unzip single file
            zip_ref.filelist[0].filename = os.path.basename(filepath)
            zip_ref.extract(zip_ref.filelist[0], os.path.dirname(filepath))
            zip_ref.close()

        # Return names of downloaded variables
        return {VARIABLE_NAME: {COMPONENT_NAME}}

    @parse_variable_component_args
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: :py:class:`xarray.Dataset()`
        """
        # Check if demand in variable and component names
        if set(variable_component_names) != (
                ensure_collection(VARIABLE_NAME, set)):
            raise ValueError("Variable names different from {'{}'}".format(
                VARIABLE_NAME))

        if variable_component_names[VARIABLE_NAME] != (
                ensure_collection(COMPONENT_NAME, set)):
            raise ValueError("Component names different from {'{}'}".format(
                COMPONENT_NAME))

        # Get places and regions
        src_place_regions, _ = self.med.geo_src.get_place_regions_for_source(
            self.name, self.AREA, **kwargs)

        # Read demand from file
        years = range(int(self.cfg['first_year']),
                      int(self.cfg['last_year']) + 1)
        da = None
        for year in years:
            # Read downloaded file
            df = self._read_downloaded_file(year, **kwargs)

            # Set UTC datetime index from 'Europe/Rome' time with DST
            start = pd.to_datetime(
                str(df.iloc[0, 0]) + '{:02d}'.format(df.iloc[0, 1] - 1),
                format='%Y%d%m%H')
            df.index = pd.date_range(
                start=start, freq='H', periods=df.shape[0],
                tz='Europe/Rome').tz_convert(None)

            # Groupby zones, summing energy
            df_zones = pd.DataFrame(
                0., index=df.index, columns=list(src_place_regions))
            for place_name, region_names in src_place_regions.items():
                df_zones[place_name] = df[region_names].sum('columns')

            # Select zones and convert to DataArray
            da_year = xr.DataArray(df_zones, dims=('time', 'region'))

            # Add array to record
            da = (da_year if da is None else
                  xr.concat([da, da_year], dim='time'))

        # Remove the years appearing from conversion to UTC
        first_date = '{}-01-01'.format(years[0])
        last_date = '{}-12-31'.format(years[-1])
        da = da.sel(time=slice(first_date, last_date))
        da['time'].attrs['timezone'] = 'UTC'

        # Add component dimension
        da = da.expand_dims('component').assign_coords(
            component=[COMPONENT_NAME])

        # Add demand to dataset
        ds = {VARIABLE_NAME: da}

        return ds

    def _read_downloaded_file(self, year, **kwargs):
        """Read downloaded file.

        :param year: Year for which to read file.
        :type year: int

        :returns: Data frame read from file.
        :rtype: :py:class:`pandas.DataFrame`
        """
        # Get filepath
        filepath = self._get_download_filepath(year, **kwargs)

        # Read file as xls or xlsx
        if not self.cfg.get('no_verbose'):
            self.log.info('- Year {} from {}'.format(
                year, filepath))
        try:
            filepath_ext = filepath + '.xls'
            df = pd.read_excel(filepath_ext, sheet_name=self.cfg['sheet_name'])
        except FileNotFoundError:
            filepath_ext = filepath + '.xlsx'
            df = pd.read_excel(filepath_ext, sheet_name=self.cfg['sheet_name'])

        return df

    def _get_url(self, year, **kwargs):
        """Get url.

        :param year: Year.
        :type year: int

        :returns: Filepath.
        :rtype: str

        .. note:: Dates in filepath are in UTC.
        """
        zip_filename = 'Anno{}.zip'.format(year)
        url = os.path.join(self.cfg['host'], zip_filename)

        return url

    def _get_download_filepath(self, year, **kwargs):
        """Get download filepath without extension.

        :param year: Year.
        :type year: int

        :returns: Filepath.
        :rtype: str

        .. note:: Dates in filepath are in UTC.
        """
        # Define filename
        filename = '{}_{}_{}'.format(
            self.DEFAULT_SRC_NAME, VARIABLE_NAME, year)

        # Get filepath
        src_dir = self.med.cfg.get_external_data_directory(self)
        filepath = os.path.join(src_dir, filename)

        return filepath

    def get_postfix(self, **kwargs):
        """Get postfix for GSE data source.

        returns: Postfix.
        rtype: str
        """
        postfix = '_{}-{}'.format(self.cfg['first_year'],
                                  self.cfg['last_year'])

        return postfix



"""Gridded-data base definitions."""
import os
from copy import deepcopy
from abc import ABC, abstractmethod
import numpy as np
import xarray as xr
from scipy.spatial import cKDTree
from .data_source import DataSourceLoaderBase


class GriddedDataSourceBase(DataSourceLoaderBase, ABC):
    """Gridded data-source abstract base class. Requires :py:meth:`get_grid`
    method to be implemented."""

    def __init__(self, med, name, cfg=None, **kwargs):
        """Build setting gridded to `True`.

        :param med: Mediator.
        :param name: Data-source name. Default is `None`.
        :param cfg: Data-source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        #: Geographic dimensions. Default values.
        self.dims = ['lat', 'lon']

        #: Mask.
        self.mask = None

        #: Stacked mask.
        self.stacked_mask = None

        #: Points in area.
        self.points_in_area = None

        #: Stacked flags for grid points in area.
        self._is_in_stack = None

        super(GriddedDataSourceBase, self).__init__(
            med, name, cfg=cfg, **kwargs)

        # Gridded dataset
        self.gridded = True

        # Add get_mask task
        self.task_mng['get_mask'] = True

    @abstractmethod
    def get_grid(self, *args, **kwargs):
        """Return data-source grid."""
        raise NotImplementedError

    def get_mask(self, **kwargs):
        """Get mask from :py:attr:`geo_src` for the given gridded data source
        and store it in :py:attr:`mask` data-source member.
        """
        # Get the mask
        if self.task_mng.get('get_mask'):
            # Make the mask
            if not self.cfg.get('no_verbose'):
                self.log.info('Making {} mask for {}'.format(
                    self.med.geo_src.name, self.name))
            kwargs_new = deepcopy(kwargs)
            kwargs_new['data_src'] = self
            self.mask = self.med.geo_src.make_mask(**kwargs_new)

            # Save the mask
            self.write_mask(**kwargs_new)

            # Update task manager
            self.task_mng['get_mask'] = False
        else:
            # Read mask
            if not self.cfg.get('no_verbose'):
                self.log.info('{} mask for {} already made'.format(
                    self.med.geo_src.name, self.name))
            self.read_mask(**kwargs)

    def get_stacked_mask(self, **kwargs):
        """Get stacked mask for area from :py:attr:`geo_src`
        for the given gridded data source and store it in
        :py:attr:`stacked_mask` data-source member.

        .. seealso:: :py:meth:_get_points_in_area
        """
        if self.stacked_mask is None:
            # Get mask
            self.get_mask(**kwargs)

            # Get points in area
            self._get_points_in_area(**kwargs)

            # Store stacked mask
            self.stacked_mask = self.mask.stack(stacked_dim=self.dims)[
                {'stacked_dim': self._is_in_stack}]

    def get_grid_postfix(self, **kwargs):
        """Return empty grid postfix string.

        returns: Grid postfix.
        rtype: str
        """
        return ''

    def get_mask_postfix(self, **kwargs):
        """Get mask postfix for data source.

        :returns: Postfix
        :rtype: str
        """
        grid_postfix = self.get_grid_postfix(**kwargs)
        geo_postfix = self.med.geo_src.get_data_postfix(**kwargs)
        postfix = '_{}{}_{}{}'.format(
            self.med.geo_src.name, geo_postfix, self.name, grid_postfix)

        return postfix

    def get_mask_path(self, makedirs=True, **kwargs):
        """Get mask file path for data source.

        :param makedirs: Make directories if needed. Default is `True`.
        :type makedirs: bool

        :returns: Filepath.
        :rtype: str
        """
        file_dir = self.med.cfg.get_project_data_directory(
            self, makedirs=makedirs, **kwargs)
        filename = 'mask{}'.format(self.get_mask_postfix(**kwargs))
        filepath = os.path.join(file_dir, filename)

        return filepath

    def read_mask(self, **kwargs):
        """Default implementation: read mask dataset as
        :py:class:`xarray.Dataset`.
        """
        filepath = '{}.nc'.format(self.get_mask_path(makedirs=False, **kwargs))
        if not self.cfg.get('no_verbose'):
            self.log.info('Reading {} mask for {} from {}'.format(
                self.med.geo_src.name, self.name, filepath))
        self.mask = xr.load_dataset(filepath)

    def write_mask(self, **kwargs):
        """Default implementation: write mask as
        :py:class:`xarray.Dataset` to netcdf.
        """
        filepath = '{}.nc'.format(self.get_mask_path(**kwargs))
        if not self.cfg.get('no_verbose'):
            self.log.info('Writing {} mask for {} to {}'.format(
                self.med.geo_src.name, self.name, filepath))
        self.mask.to_netcdf(filepath)

    def get_regional_mean(self, ds, **kwargs):
        """Average dataset over regions given by mask.

        :param ds: Dataset to average.
        :type ds: :py:class:`xarray.Dataset` of :py:class:`xarray.DataArray`

        :returns: Dataset containing regional means.
        :rtype: :py:class:`xarray.Dataset` of :py:class:`xarray.DataArray`
        """
        if not self.cfg.get('no_verbose'):
            self.log.info('Getting regional averages on {} grid'.format(
                self.name))

        if 'stacked_dim' in ds.coords:
            # Get stacked mask
            self.get_stacked_mask(**kwargs)
            mask = self.stacked_mask
        else:
            # Get mask
            self.get_mask(**kwargs)
            mask = self.mask

        # Get regional mean
        gp = ds.groupby(mask['mask'])

        # Get group means
        res = gp.mean(gp._group_dim, keep_attrs=True)

        # Remove unnecessary regions out of non-empty ones
        # and replace coordinates
        (filled_indices, idx1, idx2) = np.intersect1d(
            mask['region_index'].values, list(gp.groups),
            return_indices=True)
        res = res.loc[{'mask': filled_indices}]
        res = res.rename(mask='region')
        res['region'] = mask['region'].values[idx1]

        # Transpose region and time dimensions,
        # keeping other dimensions behind
        dim_list = list(res.dims)
        dim_list.remove('region')
        dim_list.insert(0, 'region')
        dim_list.remove('time')
        dim_list.insert(0, 'time')
        res = res.transpose(*dim_list)

        return res

    def get_total_mean(self, ds, **kwargs):
        """Average dataset over all grid points,
        independently of mask.

        :param ds: Dataset to average.
        :type ds: :py:class:`xarray.Dataset`

        :returns: Dataset containing mean.
        :rtype: :py:class:`xarray.Dataset`
        """
        if not self.cfg.get('no_verbose'):
            self.log.info('Getting {} total mean'.format(self.name))

        # Group dimensions may either original or stacked ones
        try:
            res = ds.mean(self.dims, keep_attrs=True)
        except ValueError:
            res = ds.mean(ds.stacked_dim, keep_attrs=True)

        return res

    def crop_area(self, ds, **kwargs):
        """Crop :py:obj:`ds` for data source
        over area covered by mask regions.

        :param ds: Dataset to crop.
        :type ds: :py:class:`xarray.Dataset`

        :returns: Cropped dataset.
        :rtype: :py:class:`xarray.Dataset`

        .. seealso:: :py:meth:get_stacked_mask
        """
        if not self.cfg.get('no_verbose'):
            self.log.info('Cropping area for {}'.format(self.name))

        # Make stacked mask
        self.get_stacked_mask(**kwargs)

        # Select region, removing other points
        res = ds.stack(stacked_dim=self.dims)[
            {'stacked_dim': self._is_in_stack.values}]

        return res

    def _get_points_in_area(self, **kwargs):
        """Get grid points in area."""
        if (self.points_in_area is None) or (self._is_in_stack is None):
            is_in = xr.zeros_like(self.mask['mask'], dtype=bool)
            for reg_idx in self.mask.region_index.astype(float).values:
                is_in |= (self.mask['mask'] == reg_idx)

            # Add points in area mask to regional mask dataset
            self.points_in_area = is_in

            self._is_in_stack = self.points_in_area.stack(
                stacked_dim=self.dims)


def get_geodetic_crs(cube):
    """Get geodetic Coordinate Reference System (CRS) from arbitrary system.

    :param cube: Iris cube from which to manage grids associated with
      dataset.
    :type cube: :py:class:`iris.Cube`

    :returns: Longitudes and latitudes, source CRS and geodetic CRS.
    :rtype: :py:class:`tuple` containing a :py:class:`list` and two
      CRSs.

    .. seealso:: :py:func:`get_geodetic_array`
    """
    # Get the coordinate system
    cs = cube.coord_system()

    # Get coordinates
    coords = [cube.coord('grid_longitude'), cube.coord('grid_latitude')]

    # Check if coordinate system is defined
    if cs is None:
        return coords, None, None

    # Get source CRS
    src_crs = cs.as_cartopy_crs()

    # Get geodetic CRS
    dst_crs = src_crs if src_crs.is_geodetic() else src_crs.as_geodetic()

    return coords, src_crs, dst_crs


def get_geodetic_array(ds, coords, src_crs, dst_crs):
    """ Return an array from a file making sure that geodetic longitudes
    `lon` and latitudes `lat` are included.

    :param ds: Original dataset.
    :type ds: :py:class:`xarray.Dataset`

    :returns: Geodetic array.
    :rtype: :py:class:`xarray.Dataset`

    .. seealso:: :py:func:`get_geodetic_crs`
    """
    if ('lat' not in ds.coords) or ('lon' not in ds.coords):
        # Get source grid
        x, y = coords[0].points, coords[1].points
        if x.shape != y.shape:
            x, y = np.meshgrid(x, y)

        # Convert coordinates to geodetic
        dst_coords = dst_crs.transform_points(src_crs, x, y)

        # Add the coordinates to the array
        coords_names = [c.variable_name for c in coords]
        src_coords = [(cn, ds.coords[cn]) for cn in coords_names[::-1]]
        ds.coords['lon'] = xr.DataArray(dst_coords[:, :, 0],
                                        coords=src_coords)
        ds.coords['lat'] = xr.DataArray(dst_coords[:, :, 1],
                                        coords=src_coords)

    return ds


def get_nearest_neighbor(
        ds, ref_lat, ref_lon, lat_label='lat', lon_label='lon',
        nn_tree=None, drop=False, **kwargs):
    """Evaluate dataset at point nearest to the given point.

    :param ds: Dataset to evaluate.
    :param ref_lat: Reference-point latitude.
    :param ref_lon: Reference-point longitude.
    :param lat_label: Label of latitude coordinate in dataset.
    :param lon_label: Label of longitude coordinate in dataset.
    :param nn_tree: Nearest-neighbor tree. Default is `None`,
      in which case it is computed for the dataset grid.
    :param drop: Whether to drop horizontal dimensions. Default is `False`.
    :type ds: :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray`
    :type ref_lat: float
    :type ref_lon: float
    :type lat_label: str
    :type lon_label: str
    :type nn_tree: :py:class:`scipy.spatial.KDTree`
    :type drop: bool

    :returns: Selected data, neirest neighbor index and distance to reference
      point.
    :rtype: :py:class:`tuple` of :py:class:`xarray.Dataset`
      or :py:class:`xarray.DataArray`, :py:class:`int`, :py:class:`float`
    """
    if nn_tree is None:
        # Get dataset latitude and longitudes
        LAT, LON = ds[lat_label].values, ds[lon_label].values
        if len(LAT.shape) == 1:
            LAT, LON = np.meshgrid(LAT, LON)
        lat, lon = LAT.flatten(), LON.flatten()
        ds_latlon = np.array([lat, lon]).T

        # Create nearest neighbor tree
        nn_tree = cKDTree(ds_latlon)

    # Get nearest neighbor
    ref_latlon = np.array([ref_lat, ref_lon])
    nn_dist, nn_idx = nn_tree.query(ref_latlon)

    # Select data using where in case lat/lon not coordinates
    nn_latlon = ds_latlon[nn_idx]
    ds_nn = ds.where((ds[lat_label] == nn_latlon[0]) &
                     (ds[lon_label] == nn_latlon[1]), drop=True)

    # Squeeze horizontal coordinates
    ds_nn = ds_nn.squeeze(drop=drop)

    ds_nn.attrs['nearest_neighbor_index'] = nn_idx
    ds_nn.attrs['nearest_neighbor_distance'] = nn_dist

    return ds_nn



"""EEA API."""
from ..geo import DefaultMaskAPI, DefaultMaskMaker


class DataSource(DefaultMaskAPI, DefaultMaskMaker):
    #: Default source name.
    DEFAULT_SRC_NAME = 'eea'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def get_url(self, sub_region=None, **kwargs):
        """Get URL of geographical data.

        :param sub_region: Secondary region to select. If `None`,
          the whole region is selected. Default is `None`.
        :type sub_region: str

        :returns: URL pointing to the geographical data.
        :rtype: str
        """
        sub_region = kwargs.get('sub_region', self.AREA)

        url = '{}/{}/{}{}'.format(self.cfg['host'], self.cfg['path'],
                                  sub_region.lower(), self.cfg['postfix'])

        return url

    def get_filename(self, sub_region=None, **kwargs):
        """Get downloaded-file name.

        :param sub_region: Secondary region to select. If `None`,
          the whole region is selected. Default is `None`.
        :type sub_region: str

        :returns: Filename.
        :rtype: str
        """
        return sub_region.lower()[:2] + '_' + self.cfg['resolution']
